@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 """
-Standard parameter reader and writer functions as well as output reader functions,
-including substitution of tags #JA????# or names in files.
+Standard parameter reader and writer functions as well as output reader
+functions, including substitution of tags #JA????# or names in files.
 
 Parameter files can be written in some arbitrary standard formats.
 Or a prepared parameter file can be read and parameter values replaced.
@@ -16,26 +16,42 @@ Released under the MIT License; see LICENSE file for details.
 
 History:
 
-* Written Nov 2016 by Matthias Cuntz (mc (at) macu (dot) de), including standard readers and writers
-* Added substitution function for #JA????# in prepared parameter files (sub_ja_params_files), Jan 2018, Matthias Cuntz
+* Written Nov 2016 by Matthias Cuntz (mc (at) macu (dot) de),
+  including standard readers and writers
+* Added substitution function for #JA????# in prepared parameter files
+  (sub_ja_params_files), Jan 2018, Matthias Cuntz
 * Use .pid as suffix for multiprocessing, Feb 2018, Matthias Cuntz
-* Return ndarrays instead of lists and line IDs in standard_parameter_reader, Mar 2018, Matthias Cuntz
-* Added substitution functions for arbitrary names in parameter files (sub_names_params_files*), Mar 2018, Matthias Cuntz
-* Allow substitution in several parameter files (_msub_files), Apr 2018, Matthias Cuntz
+* Return ndarrays instead of lists and line IDs in standard_parameter_reader,
+  Mar 2018, Matthias Cuntz
+* Added substitution functions for arbitrary names in parameter files
+  (sub_names_params_files*), Mar 2018, Matthias Cuntz
+* Allow substitution in several parameter files (_msub_files), Apr 2018,
+  Matthias Cuntz
 * Changed to Sphinx docstring and numpydoc, Dec 2019, Matthias Cuntz
-* Renamed standard_parameter_reader/_writer to standard_parameter_reader_bounds_mask/standard_parameter_writer_bounds_mask, Jan 2020, Matthias Cuntz
-* New standard_parameter_reader/_writer that read/write single values per parameter, Jan 2020, Matthias Cuntz
-* Swapped names and params in call to sub_names_params_files* to be compatible with new generic exe_wrapper, Jan 2020, Matthias Cuntz
-* Call standard_parameter_writer with 2 or 3 arguments, i.e. pid given or not, Jan 2020, Matthias Cuntz
-* Make all strings raw strings in sub_names_params_files_* routines to deal with regular expressions, Jan 2020, Matthias Cuntz
-* Keep formatting of names and spaces with sub_names_params functions; close input file before raising error, Feb 2020, Matthias Cuntz
+* Renamed standard_parameter_reader/_writer to
+  standard_parameter_reader_bounds_mask/standard_parameter_writer_bounds_mask,
+  Jan 2020, Matthias Cuntz
+* New standard_parameter_reader/_writer that read/write single values
+  per parameter, Jan 2020, Matthias Cuntz
+* Swapped names and params in call to sub_names_params_files* to be compatible
+  with new generic exe_wrapper, Jan 2020, Matthias Cuntz
+* Call standard_parameter_writer with 2 or 3 arguments, i.e. pid given or not,
+  Jan 2020, Matthias Cuntz
+* Make all strings raw strings in sub_names_params_files_* routines to deal
+  with regular expressions, Jan 2020, Matthias Cuntz
+* Keep formatting of names and spaces with sub_names_params functions;
+  close input file before raising error, Feb 2020, Matthias Cuntz
 * Prepend underscore to local function, Jun 2020, Matthias Cuntz
-* Use with statement for reading file so that they are properly closed, Jun 2020, Matthias Cuntz
+* Use with statement for reading file so that they are properly closed,
+  Jun 2020, Matthias Cuntz
 * Overwrite input files by default if no pid given, Jun 2020, Matthias Cuntz
 * Remove _files from function names, Jun 2020, Matthias Cuntz
 * pid keyword after arguments, Jun 2020, Matthias Cuntz
-* Rename all substitution functions sub_* to start with sub_params_*, Jun 2020, Matthias Cuntz
+* Rename all substitution functions sub_* to start with sub_params_*, Jun 2020,
+  Matthias Cuntz
 * Add pid keyword to all routines, Jun 2020, Matthias Cuntz
+* Make flake8 compliant, Dec 2020, Matthias Cuntz
+* Allow non-numeric parameters, Feb 2021, Matthias Cuntz
 
 .. moduleauthor:: Matthias Cuntz
 
@@ -60,10 +76,12 @@ import numpy as np
 
 
 __all__ = ['sub_params_ja',
-           'sub_params_names', 'sub_params_names_case', 'sub_params_names_ignorecase',
+           'sub_params_names', 'sub_params_names_case',
+           'sub_params_names_ignorecase',
            'standard_output_reader',
            'standard_parameter_reader', 'standard_parameter_writer',
-           'standard_parameter_reader_bounds_mask', 'standard_parameter_writer_bounds_mask',
+           'standard_parameter_reader_bounds_mask',
+           'standard_parameter_writer_bounds_mask',
            'standard_time_series_reader', 'standard_timeseries_reader']
 
 
@@ -96,8 +114,8 @@ def _msub(dic, text, flags=0):
         http://code.activestate.com/recipes/81330-single-pass-multiple-replace/
     and
         https://www.safaribooksonline.com/library/view/python-cookbook-2nd/0596007973/ch01s19.h
-    do not work because match.string returns the string and not the pattern so that the key
-    for the dictionary does not work anymore.
+    do not work because match.string returns the string and not the pattern so
+    that the key for the dictionary does not work anymore.
 
     History
     -------
@@ -120,32 +138,38 @@ def _msub_files(files, dic, pid=None, flags=0):
     Parameters
     ----------
     files : list
-        List with file names in which pattern replacement will be applied on each line.
+        List with file names in which pattern replacement will be applied
+        on each line.
     dic : dict
         Pattern/replacement dictionary: dic[pattern] = replacement
     pid : int, optional
-        Output files will be input files suffixed by .pid (default: overwrite input file)
+        Output files will be input files suffixed by .pid
+        (default: overwrite input file)
     flags : int, optional
         Flags will be passed to re.sub() (default: 0)
 
     Returns
     -------
     files :
-        No return value but either changed input files or output files with names of the input
-        files suffixed by .pid, in which all occurences of all patterns were replaced.
+        No return value but either changed input files or output files with
+        names of the input files suffixed by .pid, in which all occurences of
+        all patterns were replaced.
 
     History
     -------
     Written,  Matthias Cuntz, Apr 2018
     Modified, Matthias Cuntz, Dec 2019 - Sphinx docstring
               Matthias Cuntz, Jun 2020 - use with statement for reading files
-              Matthias Cuntz, Jun 2020 - overwrite input files by default if pid not given
+              Matthias Cuntz, Jun 2020 - overwrite input files by default
+                                         if pid not given
     """
     from os.path import exists
 
     for f in files:
-        if not exists(f): raise IOError('File does not exist: '+f)
-        with open(f, 'r') as fi: tt = fi.read()
+        if not exists(f):
+            raise IOError('File does not exist: '+f)
+        with open(f, 'r') as fi:
+            tt = fi.read()
         tt = _msub(dic, tt, flags=flags)
         if pid:
             fname = f+'.'+str(pid)
@@ -185,15 +209,18 @@ def sub_params_ja(files, params, pid=None):
         ...
 
     pid : int, optional
-        Output files will be input files suffixed by .pid (default: overwrite input file)
+        Output files will be input files suffixed by .pid
+        (default: overwrite input file)
 
-        Note that the default `pid=None` might not be useful if you have carefully prepared input file(s).
+        Note that the default `pid=None` might not be useful if you have
+        carefully prepared input file(s).
 
     Returns
     -------
     None
-        No return value but either changed input files or output files with names of the input
-        files suffixed by .pid, in which all #JA????# patterns were replaced by params elements.
+        No return value but either changed input files or output files with
+        names of the input files suffixed by .pid, in which all #JA????#
+        patterns were replaced by params elements.
 
     Examples
     --------
@@ -207,17 +234,24 @@ def sub_params_ja(files, params, pid=None):
               Matthias Cuntz, Mar 2018 - use _msub
               Matthias Cuntz, Apr 2018 - use _msub_files
               Matthias Cuntz, Dec 2019 - Sphinx docstring
-              Matthias Cuntz, Jun 2020 - overwrite input files by default if pid not given
+              Matthias Cuntz, Jun 2020 - overwrite input files by default
+                                         if pid not given
               Matthias Cuntz, Jun 2020 - pid keyword after arguments
+              Matthias Cuntz, Feb 2021 - allow non-numeric parameter
     """
     # assert list of files
-    if isinstance(files, str): files = [files]
+    if isinstance(files, str):
+        files = [files]
 
     # make dict for replacement
     dd = {}
     for i, p in enumerate(params):
         k = "#JA{:04d}#".format(i)
-        dd[k] = "{:.14e}".format(params[i])
+        try:
+            repl = "{:.14e}".format(p)
+        except ValueError:
+            repl = "{}".format(p)
+        dd[k] = repl
 
     # replace in each file
     _msub_files(files, dd, pid)
@@ -239,7 +273,8 @@ def sub_params_names_case(files, params, names, pid=None):
     Parameters
     ----------
     files : list
-        List with file in which values of `names` will be replaced with `params`.
+        List with file in which values of `names` will be replaced with
+        `params`.
     params : iterable
         Parameter values to be given to variables on the right of = sign
 
@@ -252,18 +287,21 @@ def sub_params_names_case(files, params, names, pid=None):
     names : iterable
         Variable names on left of = sign in files
     pid : int, optional
-        Output files will be input files suffixed by .pid (default: overwrite input file)
+        Output files will be input files suffixed by .pid
+        (default: overwrite input file)
 
     Returns
     -------
     None
-        No return value but either changed input files or output files with names of the
-        input files suffixed by .pid, in which all variables given in names are assigned
-        the values given in params.
+        No return value but either changed input files or output files with
+        names of the input files suffixed by .pid, in which all variables given
+        in names are assigned the values given in params.
 
     Examples
     --------
-    >>> sub_params_names_case([file1, file2], [0, 1, 2, 3], ['param1', 'param2', 'param3', 'param4'], pid=1234)
+    >>> sub_params_names_case([file1, file2], [0, 1, 2, 3],
+    ...                       ['param1', 'param2', 'param3', 'param4'],
+    ...                       pid=1234)
 
 
     History
@@ -272,21 +310,29 @@ def sub_params_names_case(files, params, names, pid=None):
     Modified, Matthias Cuntz, Apr 2018 - use _msub_files
               Matthias Cuntz, Dec 2019 - Sphinx docstring
               Matthias Cuntz, Jan 2020 - swap names and params in argument list
-              Matthias Cuntz, Jan 2020 - make all raw strings for regular expressions
+              Matthias Cuntz, Jan 2020 - make all raw strings for regular
+                                         expressions
               Matthias Cuntz, Feb 2020 - keep formatting of names and spaces
               Matthias Cuntz, Jun 2020 - pid keyword after arguments
+              Matthias Cuntz, Feb 2021 - allow non-numeric parameter
     """
     # assert list of files
-    if isinstance(files, str): files = [files]
+    if isinstance(files, str):
+        files = [files]
 
     # make dict for _msub with dict[pattern] = replacement
     dd = {}
     for i, p in enumerate(params):
-        nep = r"("+names[i]+r"\s*)=\s*[a-zA-Z0-9_.+-]*" # name = value
-        k = r"^(\s*)"+nep                               # beginning of line
-        dd[k] = r"\1\2= {:.14e}".format(params[i])      # replacement using substitutions \1, \2, ...
-        k = r"(\n+\s*)"+nep                             # after newline
-        dd[k] = r"\1\2= {:.14e}".format(params[i])
+        try:
+            repl = r"\1\2= {:.14e}".format(p)
+        except ValueError:
+            repl = r"\1\2= {}".format(p)
+        nep = r"("+names[i]+r"\s*)=\s*[a-zA-Z0-9_.+-]*"  # name = value
+        k = r"^(\s*)"+nep                                # beginning of line
+        dd[k] = repl                                     # replacement using
+                                                         # substitutions \1, \2
+        k = r"(\n+\s*)"+nep                              # after newline
+        dd[k] = repl
 
     # replace in each file
     _msub_files(files, dd, pid)
@@ -305,7 +351,8 @@ def sub_params_names_ignorecase(files, params, names, pid=None):
     Parameters
     ----------
     files : list
-        List with file in which values of `names` will be replaced with `params`.
+        List with file in which values of `names` will be replaced with
+        `params`.
     params : iterable
         Parameter values to be given to variables on the right of = sign
 
@@ -318,18 +365,21 @@ def sub_params_names_ignorecase(files, params, names, pid=None):
     names : iterable
         Variable names on left of = sign in files
     pid : int, optional
-        Output files will be input files suffixed by .pid (default: overwrite input file)
+        Output files will be input files suffixed by .pid
+        (default: overwrite input file)
 
     Returns
     -------
     None
-        No return value but either changed input files or output files with names of the
-        input files suffixed by .pid, in which all variables given in names are assigned
-        the values given in params.
+        No return value but either changed input files or output files with
+        names of the input files suffixed by .pid, in which all variables
+        given in names are assigned the values given in params.
 
     Examples
     --------
-    >>> sub_params_names_ignorecase([file1, file2], [0, 1, 2, 3], ['param1', 'param2', 'param3', 'param4'], pid=1234)
+    >>> sub_params_names_ignorecase([file1, file2], [0, 1, 2, 3],
+    ...                             ['param1', 'param2', 'param3', 'param4'],
+    ...                             pid=1234)
 
 
     History
@@ -338,21 +388,29 @@ def sub_params_names_ignorecase(files, params, names, pid=None):
     Modified, Matthias Cuntz, Apr 2018 - use _msub_files
               Matthias Cuntz, Dec 2019 - Sphinx docstring
               Matthias Cuntz, Jan 2020 - swap names and params in argument list
-              Matthias Cuntz, Jan 2020 - make all raw strings for regular expressions
+              Matthias Cuntz, Jan 2020 - make all raw strings for regular
+                                         expressions
               Matthias Cuntz, Feb 2020 - keep formatting of names and spaces
               Matthias Cuntz, Jun 2020 - pid keyword after arguments
+              Matthias Cuntz, Feb 2021 - allow non-numeric parameter
     """
     # assert list of files
-    if isinstance(files, str): files = [files]
+    if isinstance(files, str):
+        files = [files]
 
     # make dict for _msub with dict[pattern] = replacement
     dd = {}
     for i, p in enumerate(params):
-        nep = r"("+names[i]+r"\s*)=\s*[a-zA-Z0-9_.+-]*" # name = value
-        k = r"^(\s*)"+nep                               # beginning of line
-        dd[k] = r"\1\2= {:.14e}".format(params[i])      # replacement using substitutions \1, \2, ...
-        k = r"(\n+\s*)"+nep                             # after newline
-        dd[k] = r"\1\2= {:.14e}".format(params[i])
+        try:
+            repl = r"\1\2= {:.14e}".format(p)
+        except ValueError:
+            repl = r"\1\2= {}".format(p)
+        nep = r"("+names[i]+r"\s*)=\s*[a-zA-Z0-9_.+-]*"  # name = value
+        k = r"^(\s*)"+nep                                # beginning of line
+        dd[k] = repl                                     # replacement using
+                                                         # substitutions \1, \2
+        k = r"(\n+\s*)"+nep                              # after newline
+        dd[k] = repl
 
     # replace in each file
     _msub_files(files, dd, pid, flags=re.I)
@@ -374,8 +432,8 @@ def standard_output_reader(filename, pid=None):
     """
     Standard output reader.
 
-    The standard output reader (if outputreader=None) reads a single value from a file
-    without header, comment line or similar.
+    The standard output reader (if outputreader=None) reads a single value
+    from a file without header, comment line or similar.
 
     That means for example:
 
@@ -472,7 +530,8 @@ def standard_parameter_reader(filename, pid=None):
     params = []
     with open(fname, 'r') as ff:
         for line in ff:
-            if line.startswith('#'): continue
+            if line.startswith('#'):
+                continue
             params.append(np.float(line.strip()))
 
     return np.array(params, dtype=np.float)
@@ -527,8 +586,10 @@ def standard_parameter_writer(filename, params, pid=None):
     History
     -------
     Written,  Matthias Cuntz, Jan 2020
-    Modified, Matthias Cuntz, Jan 2020 - call with 2 or 3 arguments, i.e. pid given or not
+    Modified, Matthias Cuntz, Jan 2020 - call with 2 or 3 arguments,
+                                         i.e. pid given or not
               Matthias Cuntz, Jun 2020 - pid keyword after arguments
+              Matthias Cuntz, Feb 2021 - allow non-numeric parameter
     """
     # Existing file will be overwritten
     if pid:
@@ -537,7 +598,11 @@ def standard_parameter_writer(filename, params, pid=None):
         ofile = filename
     with open(ofile, 'w') as ff:
         for pp in params:
-            dstr = '{:.14e}'.format(pp)
+            try:
+                repl = '{:.14e}'.format(pp)
+            except ValueError:
+                repl = '{}'.format(pp)
+            dstr = repl
             print(dstr, file=ff)
 
     return
@@ -596,7 +661,8 @@ def standard_parameter_reader_bounds_mask(filename, pid=None):
 
     Examples
     --------
-    >>> ids, params, pmin, pmax, pmask = standard_parameter_reader_bounds_mask(paramfile)
+    >>> ids, params, pmin, pmax, pmask = \
+    ...     standard_parameter_reader_bounds_mask(paramfile)
 
 
     History
@@ -605,7 +671,8 @@ def standard_parameter_reader_bounds_mask(filename, pid=None):
     Modified, Matthias Cuntz, Mar 2018 - return ids
                                        - return numpy.arrays
               Matthias Cuntz, Dec 2019 - Sphinx docstring
-              Matthias Cuntz, Jan 2020 - renamed from standard_parameter_reader to standard_parameter_reader_bounds_mask
+              Matthias Cuntz, Jan 2020 - renamed from standard_parameter_reader
+                  to standard_parameter_reader_bounds_mask
               Matthias Cuntz, Feb 2020 - close file before raising error
               Matthias Cuntz, Jun 2020 - pid keyword
     """
@@ -620,9 +687,10 @@ def standard_parameter_reader_bounds_mask(filename, pid=None):
         pfile = filename
     with open(pfile, 'r') as ff:
         for line in ff:
-            l = line.strip()
-            if l.startswith('#'): continue
-            el = l.split()
+            ll = line.strip()
+            if ll.startswith('#'):
+                continue
+            el = ll.split()
             if len(el) != 5:
                 raise IOError('Line has no 5 columns for parameter: '+line)
             ids.append(el[0])
@@ -631,22 +699,28 @@ def standard_parameter_reader_bounds_mask(filename, pid=None):
             pmax.append(np.float(el[3]))
             pmask.append(int(el[4]))
 
-    return [ids, np.array(params, dtype=np.float), np.array(pmin, dtype=np.float),
-            np.array(pmax, dtype=np.float), np.array(pmask, dtype=np.bool)]
+    return [ids,
+            np.array(params, dtype=np.float),
+            np.array(pmin, dtype=np.float),
+            np.array(pmax, dtype=np.float),
+            np.array(pmask, dtype=np.bool)]
 
 
 # ------------------------------------------------------------------------------
 
 
-def standard_parameter_writer_bounds_mask(filename, params, pmin, pmax, mask, pid=None):
+def standard_parameter_writer_bounds_mask(filename, params, pmin, pmax, mask,
+                                          pid=None):
     """
     Standard parameter writer with parameter bounds and mask.
 
     The standard parameter writer writes a space separated file containing
-    1 header line (# value min max mask) plus 1 line per parameter with the following columns:
+    1 header line (# value min max mask) plus 1 line per parameter with the
+    following columns:
 
-    consecutive parameter number, current parameter value, minimum parameter value,
-    maximum parameter value, parameter mask (1: include, 0: exclude).
+    consecutive parameter number, current parameter value,
+    minimum parameter value, maximum parameter value,
+    parameter mask (1: include, 0: exclude).
 
     All values will be written in IEEE double precision: {:.14e}.
 
@@ -689,7 +763,8 @@ def standard_parameter_writer_bounds_mask(filename, params, pmin, pmax, mask, pi
     >>> randst = np.random.RandomState()
     >>> pid = str(randst.randint(2147483647))
     >>> params = sample_parameter(pis, pmin, pmax, pmask)
-    >>> standard_parameter_writer_bounds_mask(paramfile, params, pmin, pmax, pmask, pid)
+    >>> standard_parameter_writer_bounds_mask(paramfile, params, pmin, pmax,
+    ...                                       pmask, pid)
 
 
     History
@@ -697,14 +772,18 @@ def standard_parameter_writer_bounds_mask(filename, params, pmin, pmax, mask, pi
     Written,  Matthias Cuntz, Nov 2016
     Modified, Matthias Cuntz, Feb 2018 - pid
               Matthias Cuntz, Dec 2019 - Sphinx docstring
-              Matthias Cuntz, Jan 2020 - renamed from standard_parameter_writer to standard_parameter_writer_bounds_mask
+              Matthias Cuntz, Jan 2020 - renamed from standard_parameter_writer
+                  to standard_parameter_writer_bounds_mask
                                        - no .pid to filename if pid is None
               Matthias Cuntz, Jun 2020 - pid keyword after arguments
     """
     # Assert correct call
-    assert len(params) == len(pmin), 'Parameter and minima do not have the same length.'
-    assert len(params) == len(pmax), 'Parameter and maxima do not have the same length.'
-    assert len(params) == len(mask), 'Parameter and mask do not have the same length.'
+    astr = 'Parameter and minima do not have the same length.'
+    assert len(params) == len(pmin), astr
+    astr = 'Parameter and maxima do not have the same length.'
+    assert len(params) == len(pmax), astr
+    astr = 'Parameter and mask do not have the same length.'
+    assert len(params) == len(mask), astr
 
     # Convert mask to integer if boolean
     pmask = [ int(i) for i in mask ]
@@ -720,7 +799,8 @@ def standard_parameter_writer_bounds_mask(filename, params, pmin, pmax, mask, pi
         print(hstr, file=ff)
         # data
         for i in range(len(params)):
-            dstr = '{:d} {:.14e} {:.14e} {:.14e} {:d}'.format(i+1, params[i], pmin[i], pmax[i], pmask[i])
+            dstr = '{:d} {:.14e} {:.14e} {:.14e} {:d}'.format(
+                i+1, params[i], pmin[i], pmax[i], pmask[i])
             print(dstr, file=ff)
 
     return
